@@ -4,13 +4,14 @@ from datetime import timedelta
 from dataclasses import dataclass
 
 import numpy as np
+from scipy import interpolate
 from PyQt5.QtGui import (QBrush, QPen, QColor, QPalette, QPainter, QPainterPath,
     QCursor)
 from PyQt5.QtWidgets import QFrame
 from PyQt5.QtCore import Qt, pyqtSignal, QTimer, QPointF, QRectF, QRect
 from slider.beatmap import Circle, Slider, Spinner
-from slider.mod import circle_radius, od_to_ms
-from circleguard import Mod, Key
+from circleguard import (Mod, Key, hitradius, hitwindows, JudgmentType,
+    KeylessCircleguard)
 
 from circlevis.clock import Timer
 from circlevis.player import Player
@@ -26,17 +27,58 @@ PEN_GRAY = QPen(QColor(75, 75, 75))
 PEN_GREY_INACTIVE = QPen(QColor(133, 125, 125))
 PEN_HIGHLIGHT = QPen(QColor(230, 212, 92))
 PEN_BLANK = QPen(QColor(0, 0, 0, 0))
+# for missed hitobjs
+PEN_RED_TINT = QPen(QColor(200, 150, 150))
+
+# for hit error bar markers and regions
+
+PEN_BLUE = QPen(QColor(93, 183, 223))
+PEN_GREEN = QPen(QColor(127, 221, 71))
+PEN_YELLOW = QPen(QColor(211, 175, 90))
+
+BRUSH_BLUE = QBrush(QColor(93, 183, 223))
+BRUSH_GREEN = QBrush(QColor(127, 221, 71))
+BRUSH_YELLOW = QBrush(QColor(211, 175, 90))
+
 
 BRUSH_WHITE = QBrush(QColor(200, 200, 200))
 BRUSH_GRAY = QBrush(QColor(100, 100, 100))
 BRUSH_DARKGRAY = QBrush(QColor(10, 10, 10))
 BRUSH_BLANK = QBrush(QColor(0, 0, 0, 0))
+# for missed hitobjs
+BRUSH_GRAY_RED_TINT = QBrush(QColor(110, 80, 80))
+
+# for judgment indicators. red, yellow, and blue respectively
+# TODO use QRadialGradient for judgment brushes to more closely mimic karthy's
+# skin, couldn't figure out how to use them though lol
+BRUSH_JUDGMENT_MISS = QBrush(QColor(200, 27, 27))
+BRUSH_JUDGMENT_50 = QBrush(QColor(24, 75, 242))
+BRUSH_JUDGMENT_100 = QBrush(QColor(36, 181, 75))
 
 GAMEPLAY_PADDING_WIDTH = 64 + 60
 GAMEPLAY_PADDING_HEIGHT = 48 + 20
 GAMEPLAY_WIDTH = 512
 GAMEPLAY_HEIGHT = 384
 
+# how high in pixels one half of the error bar should be (so the total height
+# is twice this).
+ERROR_BAR_HEIGHT = 3
+# how wide in pixels the full error bar should be (not just half of the width)
+ERROR_BAR_WIDTH = 170
+# hitobjs which were hit less time ago than this threshold in ms will have an
+# error bar marker shown to indicate the hit
+ERROR_BAR_HIT_THRESHOLD = 4000
+# width of each hit marker in pixels
+ERROR_BAR_HIT_WIDTH = 2
+ERROR_BAR_HIT_HEIGHT = 8
+
+# radius of judgment indicator circles for 50s, 100s, 300s, misses
+JUDGMENT_INDICATOR_RADIUS = 6
+
+# hitobjs which were hit less time ago than this threshold in ms will have an
+# error bar marker shown to indicate the hit
+JUDGMENT_INDICATOR_THRESHOLD = 1000
+
 SLIDER_TICKRATE = 50
 
 
@@ -73,17 +115,31 @@ class Renderer(QFrame):
 
         self.setMouseTracking(True)
 
-        # beatmap init stuff
+        # hitobjs currently on screen
         self.hitobjs_to_draw = []
+        # we actually care about more than just the hitobjs on screen when
+        # drawing error bar markers, so keep track of those hitobjs here.
+        # this will (should) be a superset of ``hitobjs_to_draw``
+        self.hitobjs_to_draw_hits_for = []
+        # and we care about *less* than the hitobjs in hitobjs_to_draw when
+        # drawing judgment indicators, since these disappear sooner. This will
+        # be a subset of ``hitobjs_to_draw``.
+        # TODO clean up this code, these lists shouldn't exist, instead hitobjs
+        # should be marked with ``draw_judgment_indicators_for`` attributes
+        # or something
+        self.hitobjs_to_draw_judgment_indicators_for = []
 
         self.use_hr = any(Mod.HR in replay.mods for replay in replays)
         self.use_ez = any(Mod.EZ in replay.mods for replay in replays)
         if beatmap:
-            self.hit_objects = beatmap.hit_objects(hard_rock=self.use_hr, easy=self.use_ez)
+            self.hit_objects = beatmap.hit_objects(hard_rock=self.use_hr,
+                easy=self.use_ez)
             self.playback_end = self.get_hit_endtime(self.hit_objects[-1])
 
             ar = beatmap.ar(hard_rock=self.use_hr, easy=self.use_ez)
-            # https://osu.ppy.sh/help/wiki/Beatmapping/Approach_rate for formulas
+            od = beatmap.od(hard_rock=self.use_hr, easy=self.use_ez)
+            cs = beatmap.cs(hard_rock=self.use_hr, easy=self.use_ez)
+            # see https://osu.ppy.sh/help/wiki/Beatmapping/Approach_rate
             if ar <= 5:
                 self.preempt = 1200 + 600 * (5 - ar) / 5
                 self.fade_in = 800 + 400 * (5 - ar) / 5
@@ -91,13 +147,20 @@ class Renderer(QFrame):
                 self.preempt = 1200 - 750 * (ar - 5) / 5
                 self.fade_in = 800 - 500 * (ar - 5) / 5
 
-            self.hitwindow = od_to_ms(beatmap.od(hard_rock=self.use_hr, easy=self.use_ez)).hit_50
+            (hitwindow_50, hitwindow_100, hitwindow_300) = hitwindows(od)
+            self.hitwindow_50 = hitwindow_50
+            self.hitwindow_100 = hitwindow_100
+            self.hitwindow_300 = hitwindow_300
 
-            self.hitcircle_radius = circle_radius(beatmap.cs(hard_rock=self.use_hr, easy=self.use_ez))
+            # how much to scale our error bar by from its 'standard' size, where
+            # each ms of error is a pixel.
+            self.error_bar_width_factor = ERROR_BAR_WIDTH / (hitwindow_50 * 2)
+
+            self.hitcircle_radius = hitradius(cs)
             # loading stuff
             self.is_loading = True
-            # not fully accurate, but good enough
             self.num_hitobjects = len(self.hit_objects)
+            # not fully accurate, but good enough
             self.num_sliders = self.num_hitobjects
             self.sliders_current = 0
             self.thread = threading.Thread(target=self.process_sliders)
@@ -119,12 +182,17 @@ class Renderer(QFrame):
         self.num_replays = len(replays)
         self.players = []
         for i, replay in enumerate(replays):
-            self.players.append(
-                Player(replay=replay,
-                       pen=QPen(QColor().fromHslF(i / self.num_replays, 0.75, 0.5)),))
-        self.playback_end = max(max(player.t) for player in self.players) if self.num_replays > 0 else self.playback_end
-        self.playback_start = min(min(player.t) for player in self.players) if self.num_replays > 0 else 0
-        # force 0 for replays with no negative frames
+            color = QColor().fromHslF(i / self.num_replays, 0.75, 0.5)
+            player = Player(replay=replay, pen=QPen(color))
+            self.players.append(player)
+
+        self.playback_start = 0
+        if self.num_replays > 0:
+            self.playback_start = min(min(player.t) for player in self.players)
+            self.playback_end = max(max(player.t) for player in self.players)
+
+        # always start at 0, unless our playback_start is negative (meaning we
+        # have negative frames)
         self.playback_start = min(self.playback_start, 0)
 
         # if our hitobjs are hard_rock versions, flip any player *without* hr
@@ -143,7 +211,8 @@ class Renderer(QFrame):
         # render stuff
         self.timer = QTimer(self)
         self.timer.timeout.connect(self.next_frame_from_timer)
-        self.timer.start(1000/60) # 62 fps (1000ms/60frames but the result can only be a integer)
+        # 62 fps (1000ms / 60frames but the result can only be a integer)
+        self.timer.start(1000/60)
 
         # black background
         pal = QPalette()
@@ -160,9 +229,32 @@ class Renderer(QFrame):
         # how many frames for each replay to draw on screen at a time
         self.num_frames_on_screen = 15
         self.only_color_keydowns = False
+        self.should_draw_hit_error_bar = True
+        # TODO expose this as a setting somewhere? it's not toggleable anywhere
+        # currently
+        self.should_draw_judgment_indicators = True
 
         self.next_frame()
 
+        self.hitobj_to_judgments = {}
+        if self.num_replays == 1:
+            r = replays[0]
+            cg = KeylessCircleguard()
+            self.judgments = cg.judgments(r, beatmap=self.beatmap)
+            # associate each hitobject with a judgment. Only hitobjs which are
+            # spinners won't be associated in this mapping (since core doesn't
+            # generate a judgment for them yet)
+            for judgment in self.judgments:
+                # use the hitobj time as our key. This will work fine for ranked
+                # maps (no two hitobjs can be placed at the same time) but may
+                # break for aspire, loved, or crazy graveyarded maps. Needs
+                # testing.
+                # A way around this is to simply store the slider hitobj in
+                # circlecore's hitobjs, or convert core to using slider's
+                # hitobjs again (or make them subclasses, or some such)
+                self.hitobj_to_judgments[judgment.hitobject.t] = judgment
+
+
     def resizeEvent(self, event):
         width = event.size().width() - GAMEPLAY_PADDING_WIDTH * 2
         height = event.size().height() - GAMEPLAY_PADDING_HEIGHT * 2
@@ -178,10 +270,12 @@ class Renderer(QFrame):
             self.x_offset = (width - GAMEPLAY_WIDTH * y_scale) / 2
 
     def _x(self, position):
-        return self.x_offset + GAMEPLAY_PADDING_WIDTH + self.scaled_number(position)
+        return (self.x_offset + GAMEPLAY_PADDING_WIDTH +
+            self.scaled_number(position))
 
     def _y(self, position):
-        return self.y_offset + GAMEPLAY_PADDING_HEIGHT + self.scaled_number(position)
+        return (self.y_offset + GAMEPLAY_PADDING_HEIGHT +
+            self.scaled_number(position))
 
     def scaled_point(self, x, y):
         return QPointF(self._x(x), self._y(y))
@@ -191,9 +285,10 @@ class Renderer(QFrame):
 
     def next_frame_from_timer(self):
         """
-        Has the same effect as next_frame except if paused, where it returns. This is to allow
-        the back/forward buttons to advance frame by frame while still paused (as they connect directly to next
-        and previous frame), while still pausing the automatic timer advancement.
+        Has the same effect as next_frame except if paused, where it returns.
+        This is to allow the back/forward buttons to advance frame by frame
+        while still paused (as they connect directly to next and previous
+        frame), while still pausing the automatic timer advancement.
         """
         if self.paused:
             # ignore our paused state if we're still loading sliders, or else if
@@ -244,7 +339,10 @@ class Renderer(QFrame):
             # right side case
             if side == "right":
                 player.end_pos -= 1
-            player.start_pos = player.end_pos - self.num_frames_on_screen if player.end_pos >= self.num_frames_on_screen else 0
+
+            player.start_pos = 0
+            if player.end_pos >= self.num_frames_on_screen:
+                player.start_pos = player.end_pos - self.num_frames_on_screen
 
         if self.has_beatmap:
             self.get_hitobjects()
@@ -252,7 +350,7 @@ class Renderer(QFrame):
         self.update()
 
     def get_hitobjects(self):
-        # get current hitobjects
+        # get currently visible hitobjects
         current_time = self.clock.get_time()
         found_all = False
         # TODO optimize this by tracking our current hitobj index, this iterates
@@ -260,24 +358,31 @@ class Renderer(QFrame):
         # O(n) worst case) which can't be good for performance
         index = 0
         self.hitobjs_to_draw = []
+        self.hitobjs_to_draw_hits_for = []
+        self.hitobjs_to_draw_judgment_indicators_for = []
         while not found_all:
             current_hitobj = self.hit_objects[index]
             hit_t = current_hitobj.time.total_seconds() * 1000
-            if isinstance(current_hitobj, Slider) or isinstance(current_hitobj, Spinner):
+            if isinstance(current_hitobj, (Slider, Spinner)):
                 hit_end = self.get_hit_endtime(current_hitobj) + self.fade_in
             else:
-                hit_end = hit_t + self.hitwindow + self.fade_in
+                hit_end = hit_t + self.hitwindow_50 + self.fade_in
+            if hit_t > current_time - JUDGMENT_INDICATOR_THRESHOLD:
+                self.hitobjs_to_draw_judgment_indicators_for.append(current_hitobj)
+            if hit_t > current_time - ERROR_BAR_HIT_THRESHOLD:
+                self.hitobjs_to_draw_hits_for.append(current_hitobj)
             if hit_t - self.preempt < current_time < hit_end:
                 self.hitobjs_to_draw.append(current_hitobj)
+
             elif hit_t > current_time:
                 found_all = True
             if index == self.num_hitobjects - 1:
                 found_all = True
             index += 1
 
-    def paintEvent(self, event):
+    def paintEvent(self, _event):
         """
-        Called whenever self.update() is called. Draws all cursors and Hitobjects
+        Called whenever self.update() is called
         """
         self.painter.begin(self)
         self.painter.setRenderHint(QPainter.TextAntialiasing, True)
@@ -303,12 +408,20 @@ class Renderer(QFrame):
             self.paint_cursor(player)
         # other info
         self.painter.setPen(_pen)
+        self.paint_border()
         if self.should_paint_info:
             self.paint_info()
         if self.paint_frametime:
             self.paint_frametime_graph()
         self.painter.end()
 
+    def paint_border(self):
+        PEN_WHITE.setWidth(self.scaled_number(1))
+        self.painter.setPen(PEN_WHITE)
+        self.painter.setOpacity(0.25)
+        self.painter.drawRect(QRectF(self.scaled_point(0, 0),
+            self.scaled_point(GAMEPLAY_WIDTH, GAMEPLAY_HEIGHT)))
+
     def paint_cursor(self, player):
         """
         Draws a cursor.
@@ -327,7 +440,8 @@ class Renderer(QFrame):
         self.painter.setPen(pen)
         highlighted_pen = False
         for i in range(player.start_pos, player.end_pos):
-            highlight = any((player.t[i + 1] in self.events, player.t[i] in self.events))
+            highlight = any((player.t[i + 1] in self.events, player.t[i] in
+                self.events))
             if highlight and not highlighted_pen:
                 self.painter.setPen(PEN_HIGHLIGHT)
                 highlighted_pen = True
@@ -366,24 +480,58 @@ class Renderer(QFrame):
         self.painter.setOpacity(1)
 
     def paint_beatmap(self):
+        # draw playfield judgment indicators (yellow/green/blue circles under
+        # hitobjs) before drawing hitobjs so they don't cover hitobjs
+        # (though to be honest it doesn't make much of a difference either way)
+
+        if self.should_draw_judgment_indicators and self.num_replays == 1:
+            for hitobj in self.hitobjs_to_draw_hits_for:
+                if isinstance(hitobj, Spinner):
+                    continue
+
+                judgment = self.hitobj_to_judgments[self.get_hit_time(hitobj)]
+
+                if judgment.type is JudgmentType.Miss:
+                    # misses don't have an intrinsic event time, so just use
+                    # their hitobj's time
+                    t = judgment.hitobject.time
+                else:
+                    t = judgment.time
+
+                # don't draw hits that haven't happened yet
+                if t <= self.clock.get_time():
+                    self.draw_judgment_indicator(hitobj, judgment)
+            self.painter.setBrush(BRUSH_BLANK)
+
         for hitobj in self.hitobjs_to_draw[::-1]:
             self.draw_hitobject(hitobj)
 
+        # only draw hit error bars if there's only one replay
+        if self.should_draw_hit_error_bar and self.num_replays == 1:
+            self.draw_hit_error_bar()
+
+            for hitobj in self.hitobjs_to_draw_hits_for:
+                # core doesn't calculate judgmnets for spinners yet, TODO
+                # implement this when core does
+                if isinstance(hitobj, Spinner):
+                    continue
+
+                judgment = self.hitobj_to_judgments[self.get_hit_time(hitobj)]
+                # don't draw any judgment bars for misses
+                if judgment.type is JudgmentType.Miss:
+                    continue
+                # don't draw hits that haven't happened yet
+                if judgment.t <= self.clock.get_time():
+                    self.draw_hit(hitobj, judgment)
+
     def paint_info(self):
         """
         Draws various info about the replays in the upper left corner.
-
-        Args:
-           QPainter painter: The painter.
         """
         # our current y coordinate for drawing info. Modified throughout this
         # function
         y = 15
 
-        PEN_WHITE.setWidth(self.scaled_number(1))
-        self.painter.setPen(PEN_WHITE)
-        self.painter.setOpacity(0.25)
-        self.painter.drawRect(QRectF(self.scaled_point(0, 0), self.scaled_point(GAMEPLAY_WIDTH, GAMEPLAY_HEIGHT)))
         PEN_WHITE.setWidth(1)
         self.painter.setPen(PEN_WHITE)
         self.painter.setOpacity(1)
@@ -441,22 +589,26 @@ class Renderer(QFrame):
                 pen = player.pen
                 self.painter.setPen(PEN_BLANK)
                 self.painter.setBrush(QBrush(pen.color()))
-                _set_opacity(1 if Key.M1 in Key(int(player.k[player.end_pos])) else 0.3)
+                keys = Key(int(player.k[player.end_pos]))
+                _set_opacity(1 if Key.M1 in keys else 0.3)
                 self.painter.drawRect(5, y - 9, 10, 10)
-                _set_opacity(1 if Key.M2 in Key(int(player.k[player.end_pos])) else 0.3)
+                _set_opacity(1 if Key.M2 in keys else 0.3)
                 self.painter.drawRect(18, y - 9, 10, 10)
                 _set_opacity(1)
                 self.painter.setPen(pen)
                 info_text = (f"{player.username} {player.mods.short_name()}: "
-                    f"{player.xy[player.end_pos][0]:.2f}, {player.xy[player.end_pos][1]:.2f}")
+                    f"{player.xy[player.end_pos][0]:.2f}, "
+                    f"{player.xy[player.end_pos][1]:.2f}")
                 self.painter.drawText(31, y, info_text)
                 # not sure why we need to do ``y - 9`` instead of 9 here,
                 # our ``drawText`` call is perfectly happy to accept ``y`` but
                 # we need to pass ``y - 9`` to our ``drawRect`` calls...maybe 9
                 # was a manually determined number that causes the text to align
                 # with the drawn boxes?
-                info_pos = self.painter.boundingRect(5, y - 9, 0, 0, 0, info_text)
-                info_pos = Rect(info_pos.x(), info_pos.y(), info_pos.width(), info_pos.height())
+                info_pos = self.painter.boundingRect(5, y - 9, 0, 0, 0,
+                    info_text)
+                info_pos = Rect(info_pos.x(), info_pos.y(), info_pos.width(),
+                    info_pos.height())
                 # unfortunately the rects overlap if we don't make this manual
                 # adjustment; would like to figure out why but this works for
                 # now.
@@ -476,7 +628,8 @@ class Renderer(QFrame):
                     distance = math.sqrt(((p1.xy[p1.end_pos][0] - p2.xy[p2.end_pos][0]) ** 2) +
                                          ((p1.xy[p1.end_pos][1] - p2.xy[p2.end_pos][1]) ** 2))
                     self.painter.drawText(5, y, f"{int(distance)}px apart")
-                except IndexError: # Edge case where we only have data from one cursor
+                except IndexError:
+                    # we may only have data from one cursor at the moment
                     pass
 
             if self.num_replays == 1 and self.has_beatmap:
@@ -486,7 +639,8 @@ class Renderer(QFrame):
                 closest_hitobj = self.beatmap.closest_hitobject(current_t)
                 if self.use_hr:
                     closest_hitobj = closest_hitobj.hard_rock
-                distance = self.distance_between(player.xy[player.end_pos], closest_hitobj)
+                distance = self.distance_between(player.xy[player.end_pos],
+                    closest_hitobj)
 
                 # show "x px inside hitobj" instead of a negative distance
                 inside = False
@@ -494,7 +648,9 @@ class Renderer(QFrame):
                     inside = True
                     distance = abs(distance)
 
-                self.painter.drawText(5, y, f"{distance:0.2f}px {'inside' if inside else 'from'} closest hitobj")
+                inside_from = "inside" if inside else "from"
+                text = f"{distance:0.2f}px {inside_from} closest hitobj"
+                self.painter.drawText(5, y, text)
 
             for function in self.statistic_functions:
                 y += 13
@@ -506,7 +662,8 @@ class Renderer(QFrame):
 
     def draw_line(self, alpha, start, end, grey_out=False):
         """
-        Draws a line at the given alpha level from the start point to the end point.
+        Draws a line at the given alpha level from the start point to the end
+        point.
 
         Arguments:
             Float alpha: The alpha level (from 0.0 to 1.0) to set the line to.
@@ -520,7 +677,8 @@ class Renderer(QFrame):
             self.painter.setPen(PEN_GREY_INACTIVE)
 
         self.painter.setOpacity(alpha)
-        self.painter.drawLine(self.scaled_point(start[0], start[1]), self.scaled_point(end[0], end[1]))
+        self.painter.drawLine(self.scaled_point(start[0], start[1]),
+            self.scaled_point(end[0], end[1]))
 
         if self.raw_view and grey_out:
             self.painter.setPen(prev_pen)
@@ -565,11 +723,7 @@ class Renderer(QFrame):
 
     def draw_hitobject(self, hitobj):
         """
-        Calls corresponding functions to draw a Hitobject.
-
-        Args:
-            QPainter painter: The painter.
-            Hitobj hitobj: A Hitobject.
+        Calls the corresponding function to draw ``hitobj``.
         """
         if not self.draw_hitobjects:
             return
@@ -583,14 +737,10 @@ class Renderer(QFrame):
 
     def draw_hitcircle(self, hitobj):
         """
-        Draws Hitcircle.
-
-        Args:
-            QPainter painter: The painter.
-            Hitobj hitobj: A Hitobject.
+        Draws a circle hitobject.
         """
         current_time = self.clock.get_time()
-        fade_out = max(0, ((current_time - self.get_hit_time(hitobj)) / self.hitwindow))
+        fade_out = max(0, ((current_time - self.get_hit_time(hitobj)) / self.hitwindow_50))
         opacity = min(1, ((current_time - (self.get_hit_time(hitobj) - self.preempt)) / self.fade_in))
         opacity = max(0, min(1, opacity-fade_out))
         p = hitobj.position
@@ -601,26 +751,33 @@ class Renderer(QFrame):
         # have radius `self.hitcircle_radius`.
         r = self.scaled_number(self.hitcircle_radius - WIDTH_CIRCLE_BORDER / 2)
 
-        PEN_WHITE.setWidth(self.scaled_number(WIDTH_CIRCLE_BORDER))
+        judgment = self.hitobj_to_judgments[self.get_hit_time(hitobj)]
+        if judgment.type is JudgmentType.Miss:
+            # hitobj was missed, tint red
+            pen = PEN_RED_TINT
+            brush = BRUSH_GRAY_RED_TINT
+        else:
+            # hitobj was hit
+            pen = PEN_WHITE
+            brush = BRUSH_GRAY
+
+        pen.setWidth(self.scaled_number(WIDTH_CIRCLE_BORDER))
+        self.painter.setPen(pen)
         self.painter.setOpacity(opacity)
-        self.painter.setPen(PEN_WHITE)
-        self.painter.setBrush(BRUSH_GRAY)
+        self.painter.setBrush(brush)
+
         self.painter.drawEllipse(self.scaled_point(p.x, p.y), r, r)
         self.painter.setBrush(BRUSH_BLANK)
 
     def draw_spinner(self, hitobj):
         """
-        Draws Spinner.
-
-        Args:
-            QPainter painter: The painter.
-            Hitobj hitobj: A Hitobject.
+        Draws a spinner hitobject.
         """
         current_time = self.clock.get_time()
         if self.get_hit_endtime(hitobj) - current_time < 0:
             return
         radius = GAMEPLAY_HEIGHT / 2
-        fade_out = max(0, ((current_time - self.get_hit_endtime(hitobj)) / self.hitwindow))
+        fade_out = max(0, ((current_time - self.get_hit_endtime(hitobj)) / self.hitwindow_50))
         opacity = min(1, ((current_time - (self.get_hit_time(hitobj) - self.preempt)) / self.fade_in))
         opacity = max(0, min(1, opacity-fade_out))
         scale = min(1, (self.get_hit_endtime(hitobj) - current_time) / (self.get_hit_endtime(hitobj) - self.get_hit_time(hitobj)))
@@ -634,11 +791,7 @@ class Renderer(QFrame):
 
     def draw_approachcircle(self, hitobj):
         """
-        Draws Approachcircle.
-
-        Args:
-            QPainter painter: The painter.
-            Hitobj hitobj: A Hitobject.
+        Draws the approach circle of a circle hitobject.
         """
         if not self.draw_approach_circles:
             return
@@ -651,18 +804,22 @@ class Renderer(QFrame):
         p = hitobj.position
         r = self.scaled_number(self.hitcircle_radius * scale)
 
-        PEN_WHITE.setWidth(self.scaled_number(WIDTH_CIRCLE_BORDER / 2))
-        self.painter.setPen(PEN_WHITE)
+        judgment = self.hitobj_to_judgments[self.get_hit_time(hitobj)]
+        if judgment.type is JudgmentType.Miss:
+            # hitobj was missed, tint red
+            pen = PEN_RED_TINT
+        else:
+            # hitobj was hit
+            pen = PEN_WHITE
+
+        pen.setWidth(self.scaled_number(WIDTH_CIRCLE_BORDER / 2))
+        self.painter.setPen(pen)
         self.painter.setOpacity(opacity)
         self.painter.drawEllipse(self.scaled_point(p.x, p.y), r, r)
 
     def draw_slider(self, hitobj):
         """
-        Draws sliderbody and hitcircle & approachcircle if needed
-
-        Args:
-            QPainter painter: The painter.
-            Hitobj hitobj: A Hitobject.
+        Draws sliderbody, hitcircle, approachcircle if needed
         """
         self.draw_sliderbody(hitobj)
         self.draw_hitcircle(hitobj)
@@ -670,15 +827,11 @@ class Renderer(QFrame):
 
     def draw_sliderbody(self, hitobj):
         """
-        Draws a sliderbody using a QpainterPath.
-
-        Args:
-            QPainter painter: The painter.
-            Hitobj hitobj: A Hitobject.
+        Draws the sliderbody of a slider using a QpainterPath.
         """
 
         current_time = self.clock.get_time()
-        fade_out = max(0, ((current_time - self.get_hit_endtime(hitobj)) / self.hitwindow))
+        fade_out = max(0, ((current_time - self.get_hit_endtime(hitobj)) / self.hitwindow_50))
         opacity = min(1, ((current_time - (self.get_hit_time(hitobj) - self.preempt)) / self.fade_in))
         opacity = max(0, min(1, opacity-fade_out)) * 0.75
         p = hitobj.position
@@ -695,6 +848,120 @@ class Renderer(QFrame):
             sliderbody.lineTo(self.scaled_point(i.x, i.y))
         self.painter.drawPath(sliderbody)
 
+    def draw_hit_error_bar(self):
+        mid_x = GAMEPLAY_WIDTH / 2
+        y = GAMEPLAY_HEIGHT - ERROR_BAR_HIT_HEIGHT
+
+        # draw the center white bar
+        self.painter.setPen(PEN_WHITE)
+        pen = self.painter.pen()
+        pen.setWidth(ERROR_BAR_HIT_WIDTH)
+        self.painter.setPen(pen)
+        self.draw_line(1, [mid_x, y - ERROR_BAR_HIT_HEIGHT],
+            [mid_x, y + ERROR_BAR_HIT_HEIGHT])
+
+        # draw the three error zones as slightly transparent
+        self.painter.setOpacity(0.65)
+        self.painter.setPen(PEN_BLANK)
+
+        hw300 = self.hitwindow_300 * self.error_bar_width_factor
+        hw100 = self.hitwindow_100 * self.error_bar_width_factor
+        hw50 = self.hitwindow_50 * self.error_bar_width_factor
+
+        self.painter.setBrush(BRUSH_BLUE)
+        p1 = self.scaled_point(mid_x - hw300, y - ERROR_BAR_HEIGHT)
+        p2 = self.scaled_point(mid_x + hw300, y + ERROR_BAR_HEIGHT)
+        self.painter.drawRect(QRectF(p1, p2))
+
+        # draw two rects to avoid overlapping with hitwindow_300 in the center
+        self.painter.setBrush(BRUSH_GREEN)
+        p1 = self.scaled_point(mid_x - hw100, y - ERROR_BAR_HEIGHT)
+        p2 = self.scaled_point(mid_x - hw300, y + ERROR_BAR_HEIGHT)
+        self.painter.drawRect(QRectF(p1, p2))
+        p1 = self.scaled_point(mid_x + hw300, y - ERROR_BAR_HEIGHT)
+        p2 = self.scaled_point(mid_x + hw100, y + ERROR_BAR_HEIGHT)
+        self.painter.drawRect(QRectF(p1, p2))
+
+        self.painter.setBrush(BRUSH_YELLOW)
+        p1 = self.scaled_point(mid_x - hw50, y - ERROR_BAR_HEIGHT)
+        p2 = self.scaled_point(mid_x - hw100, y + ERROR_BAR_HEIGHT)
+        self.painter.drawRect(QRectF(p1, p2))
+        p1 = self.scaled_point(mid_x + hw100, y - ERROR_BAR_HEIGHT)
+        p2 = self.scaled_point(mid_x + hw50, y + ERROR_BAR_HEIGHT)
+        self.painter.drawRect(QRectF(p1, p2))
+
+        self.painter.setBrush(BRUSH_BLANK)
+        self.painter.setOpacity(1)
+
+    def draw_hit(self, hitobj, hit):
+        # TODO: avoid duplication in these constants between this and
+        # `draw_hit_error_bar` - maybe just extract to globals?
+        mid_x = GAMEPLAY_WIDTH / 2
+        y = GAMEPLAY_HEIGHT - ERROR_BAR_HIT_HEIGHT
+
+        if hit.type is JudgmentType.Hit300:
+            pen = PEN_BLUE
+        elif hit.type is JudgmentType.Hit100:
+            pen = PEN_GREEN
+        elif hit.type is JudgmentType.Hit50:
+            pen = PEN_YELLOW
+
+        self.painter.setPen(pen)
+        pen = self.painter.pen()
+        pen.setWidth(ERROR_BAR_HIT_WIDTH)
+        self.painter.setPen(pen)
+
+        current_time = self.clock.get_time()
+
+        # positive is a late hit, negative is an early hit
+        error = (hit.t - self.get_hit_time(hitobj)) * self.error_bar_width_factor
+        start = [mid_x + error, y - ERROR_BAR_HIT_HEIGHT]
+        end = [mid_x + error, y + ERROR_BAR_HIT_HEIGHT]
+
+        time_passed = current_time - hit.t
+        # how long in ms to display hits for before they disappear
+        time_passed = min(time_passed, ERROR_BAR_HIT_THRESHOLD)
+        # draw most recent hits as more visible (higher alpha) and old hits
+        # as less visible (lower alpha)
+        x_interp = [0, ERROR_BAR_HIT_THRESHOLD]
+        y_interp = [1, 0]
+        f = interpolate.interp1d(x_interp, y_interp)
+        alpha = f(time_passed)
+        self.draw_line(alpha, start, end)
+
+    def draw_judgment_indicator(self, hitobj, judgment):
+        if judgment.type is JudgmentType.Hit300:
+            # don't draw anything for 300s
+            return
+
+        if judgment.type is JudgmentType.Miss:
+            brush = BRUSH_JUDGMENT_MISS
+            judgment_t = judgment.hitobject.time
+        else:
+            judgment_t = judgment.time
+        if judgment.type is JudgmentType.Hit50:
+            brush = BRUSH_JUDGMENT_50
+        if judgment.type is JudgmentType.Hit100:
+            brush = BRUSH_JUDGMENT_100
+
+        self.painter.setPen(PEN_BLANK)
+        self.painter.setBrush(brush)
+
+        current_time = self.clock.get_time()
+
+        time_passed = current_time - judgment_t
+        time_passed = min(time_passed, JUDGMENT_INDICATOR_THRESHOLD)
+        x_interp = [0, JUDGMENT_INDICATOR_THRESHOLD]
+        y_interp = [1, 0]
+        f = interpolate.interp1d(x_interp, y_interp)
+        alpha = f(time_passed)
+
+        self.painter.setOpacity(alpha)
+
+        p = hitobj.position
+        r = self.scaled_number(JUDGMENT_INDICATOR_RADIUS)
+        self.painter.drawEllipse(self.scaled_point(p.x, p.y), r, r)
+
     def draw_progressbar(self, percentage):
         loading_bg = QPainterPath()
         loading_bar = QPainterPath()
@@ -711,7 +978,8 @@ class Renderer(QFrame):
         loading_bg.lineTo(self.width()/2 - 75 + 150, self.height() / 2)
 
         loading_bar.moveTo(self.width() / 2 - 75, self.height() / 2)
-        loading_bar.lineTo(self.width() / 2 - 75 + percentage * 1.5, self.height() / 2)
+        loading_bar.lineTo(self.width() / 2 - 75 + percentage * 1.5,
+            self.height() / 2)
 
         self.painter.drawPath(loading_bg)
         _pen.setColor(QColor(c.red(), c.green(), c.blue(), 255))
@@ -719,8 +987,11 @@ class Renderer(QFrame):
         self.painter.drawPath(loading_bar)
 
     def draw_loading_screen(self):
-        self.painter.drawText(self.width() / 2 - 75, self.height() / 2 - 10, "Calculating Sliders, please wait...")
-        self.draw_progressbar(int((self.sliders_current / self.num_sliders) * 100))
+        x = self.width() / 2 - 75
+        y = self.height() / 2 - 10
+        self.painter.drawText(x, y, "Calculating Sliders, please wait...")
+        progress = int((self.sliders_current / self.num_sliders) * 100)
+        self.draw_progressbar(progress)
 
     def process_sliders(self):
         for i, hitobj in enumerate(self.hit_objects):
@@ -824,10 +1095,17 @@ class Renderer(QFrame):
         return super().mousePressEvent(event)
 
     def get_hit_endtime(self, hitobj):
-        return hitobj.end_time.total_seconds() * 1000 if not isinstance(hitobj, Circle) else self.get_hit_time(hitobj)
+        if isinstance(hitobj, Circle):
+            return self.get_hit_time(hitobj)
+        t = hitobj.end_time.total_seconds() * 1000
+        return int(round(t))
 
     def get_hit_time(self, hitobj):
-        return hitobj.time.total_seconds() * 1000
+        t = hitobj.time.total_seconds() * 1000
+        # Due to floating point errors, ``t`` could actually be something
+        # like ``129824.99999999999`` or ``128705.00000000001``, so round to the
+        # nearest int.
+        return int(round(t))
 
     def pause(self):
         """
@@ -879,6 +1157,10 @@ class Renderer(QFrame):
         self.num_frames_on_screen = new_value
         self.update()
 
+    def draw_hit_error_bar_changed(self, new_value):
+        self.should_draw_hit_error_bar = new_value
+        self.update()
+
     def circle_size_mod_changed(self, new_value):
         if not self.has_beatmap:
             # cs doesn't matter to us if we don't have a beatmap (and we don't
@@ -886,7 +1168,8 @@ class Renderer(QFrame):
             return
         use_hr = new_value == "HR"
         use_ez = new_value == "EZ"
-        self.hitcircle_radius = circle_radius(self.beatmap.cs(hard_rock=use_hr, easy=use_ez))
+        cs = self.beatmap.cs(hard_rock=use_hr, easy=use_ez)
+        self.hitcircle_radius = hitradius(cs)
         self.update()
 
 
